# coding: utf-8

import json
import os
from datetime import datetime
from copy import deepcopy as _deepcopy

from ..exceptions import OutputFileAlreadyExists

try:
    xrange
except NameError:
    xrange = range


json_indent = 1


class JSONEncoder(json.JSONEncoder):
    """
    Custom JSON Encoder for Container.
    """
    encoder_map = {
        datetime: (lambda dt: dt.isoformat()),
    }

    def default(self, o):
        try:
            encoder = self.encoder_map[type(o)]
            return encoder(o)
        except KeyError:
            return str(o)


def iterize(iterable):
    """
    :param iterable: iterable objects or int or None(return None). <iterable> or <int> or <None>.
    :return: iterable objects
    """
    if iterable is None:
        return None

    if isinstance(iterable, int):
        return xrange(iterable)

    return iterable


class BaseContainer(object):
    """
    Base Container object.
    """
    def __init__(self, blueprint, iterable=()):
        """
        :param blueprint: pattern / model / container / function / else(fixed value)
        :param iterable: iterable objects or int
        :return: container instance.
        """
        self.blueprint = blueprint
        self.iterable = iterize(iterable)

        self.set_shape(
            lambda v: v
        )
        self.set_stringify(
            lambda v: json.dumps(v, indent=json_indent, cls=JSONEncoder)
        )

    def repeat(self, iterable):
        """
        :param iterable:
        :return:
        """
        self.iterable = iterize(iterable)
        return self

    def count(self):
        return len(self)

    def stringify(self, deepcopy=True):
        """
        stringify shaped self(container) object.
        :return:
        """
        shaped = self.shape(deepcopy=deepcopy)
        return self._stringify(shaped)

    def set_stringify(self, stringify):
        """
        set how to convert to a string.(default: json.dumps(self, indent=1))
        :param stringify:
        :return:
        """
        self._stringify = stringify

    def shape(self, deepcopy=True):
        """
        shape self(container) object.
        :return: shaped container object
        """
        for base in self.__class__.__bases__:
            if not base.__name__.endswith('Container'):
                obj = base(self)
                break

        if deepcopy:
            obj = _deepcopy(obj)

        return self._shape(obj)

    def set_shape(self, shape):
        """
        set how to shape self(container) object.(default: no operation)
        :return:
        """
        self._shape = shape

    def output(self, path, rewrite=False, charset='utf8', deepcopy=True):
        """
        self(container) object write to file.
        :param path: output file path.
        :param rewrite: rewrite if file path is already exists.(default: False)
        :return: None
        """
        string = self.stringify(deepcopy=deepcopy)
        if os.path.exists(path):
            if not rewrite:
                raise OutputFileAlreadyExists()

        if charset:
            import codecs
            open_method = codecs.open
            open_args = (path, 'wb', charset)
        else:
            open_method = open
            open_args = (path, 'wb')

        #with open(path, 'wb') as f:
        with open_method(*open_args) as f:
            f.write(string)

